/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:27 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


/*---------------------------------------------------------------

	Map patterns in a high-dimensional pattern space to a
	lower dimension, preserving their structure. Reference:

Sammon, J. W. "A Non-linear mapping for data structure analysis",
IEEE Transactions on Computers, Vol. C-18, No. 5, pp 401-409, 1969.

Original nomenclature of the paper used in the routine 'map_sammon()'

-----------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"

extern float Euclidian_Distance();
extern float *vector();
extern double *dbl_vector();
extern float **matrix_alloc();

extern universe *U;
extern boolean verbose;
extern boolean feat_description;
extern char **feature_desc;


static str80 buf;
static str200 str_gnu, str_gnu1;


#define MF_DEFAULT	(0.2)
#define MF_MAX		(1.0)
static float MF = MF_DEFAULT;		/* Learning rate */

#define EPSILON (0.001)

boolean map_sammon( X, N, L, Y, d, nrIter, mf )
float **X;		/* Pointers to all sample vectors */
int N;			/* Number of all samples */
int L;			/* Original dimension of pattern space */
float **Y;		/* Pointers to the mapped samples */
int d;			/* Dimension of mapped space */
int nrIter;		/* Number of iterations of gradient descent */
float mf;		/* Learning rate of gradient descent */
{
 float **Y_new = NULL;	/* copy of the samples for each pattern loop 'p' */
 float *distX = NULL;	/* mutual distances in original feature space */
 double c;		/* sum of all mutual distances */
 double err;		/* mapping error */
 float *dE = NULL, *d2E2 = NULL;
 int nd;		/* number of mutual distances */
 double *meanY = NULL;	/* centroid of mapped samples */
 int i, j, p, q, iter, m;
 float dpj_star, dpj, dpj_diff, dpj_prod, Y_diff, delta_pq;
 float dij_star, dij, dij_diff;
 int *auxRandom1 = NULL, *auxRandom2 = NULL;

 if( d >= L || d < 1 || d > L || L < 2 )
 {
   fprintf(stderr, "\nSammon map: It makes no sense to map %d-D to %d-D\n",
	N, d );
   return FALSE ;
 }
 if( nrIter < 1 || mf <= 0.0 )
 {
   fprintf(stderr,
	"\nSammon map: Invalid number of iterations %d or learning rate %f\n",
		nrIter, mf );
   return FALSE ;
 }

 /*  --- Initialization of the mapped vectors ---  */
 if( d > 1 )
 {
   for( i = 0; i < N; i++ )
   {
     /* initialize: set the mapped vector randomly */
     random_unit_vector( Y[i], d );
     /* printf("Random init Y[%3d]=",i ); showFV( d, Y[i] ); DBG;   /**/
   }
 }
 else	/* d == 1 */
 {
   /* no sense in generating 1-dim random unit vector */
   if( d != 1 ) EXIT("Invalid dimension in 'map_sammon()'");
   auxRandom1 = (int*) malloc( N * sizeof(int) ); CHKPTR(auxRandom1);
   auxRandom2 = (int*) malloc( N * sizeof(int) ); CHKPTR(auxRandom2);
   for( i = 0; i < N; i++ )
     auxRandom1[i] = i;
   shuffle( auxRandom1, auxRandom2, N );
   for( i = 0; i < N; i++ )
     Y[i][0] = (float)auxRandom2[i] / (float)N;
   FREE(auxRandom1); FREE(auxRandom2);
   /* for( i = 0; i < N; i++ )
     printf("Y[%d][0]=%f\n", i, Y[i][0] ); DBG;	/**/
 }

 nd = ( N * (N - 1) / 2 );

 /*  --- Allocation of memory ---  */
 distX = vector( nd ); if( distX == NULL ) return FALSE;
 meanY = dbl_vector( d ); if( meanY == NULL )
   { FREE(distX); return FALSE; }
 dE = vector( d ); if( dE == NULL )
   { FREE(meanY); FREE(distX); return FALSE; }
 d2E2 = vector( d ); if( d2E2 == NULL )
   { FREE(dE); FREE(meanY); FREE(distX); return FALSE; }
 Y_new = matrix_alloc( N, d );
 if( Y_new == NULL )
   { FREE(d2E2); FREE(dE); FREE(meanY); FREE(distX); return FALSE; }


 printf("Calculating %d mutual distances among %d samples...\r", nd, N );
 fflush(stdout);
 m = 0; c = 0.0;
 for (j = 1; j < N; j++)
 {
   for (i = 0; i < j; i++)
   {
     distX[m] = Euclidian_Distance( X[i], X[j], L );
     c += (double)distX[m];
     /* showFV( L, X[i] ); showFV( L, X[j] );	/**/
     /* printf("\tm=%3d  Mutual[%d][%d] = %7.5f", m, i, j, distX[m] ); NL; /**/
     if( distX[m] == 0.0 )
     {
       BLANK_LINE;
       printf("WARNING: Zero distance between samples %d and %d\n", j+1, i+1 );
       printf("\tSetting distance to an epsilon: %f\n", EPSILON );
       distX[m] = EPSILON;
     }
     m++;
   }
 }
 BLANK_LINE;

 for( iter = 0; iter < nrIter; iter++ )
 {
   printf("Learning steps to go:    %6d   ", nrIter-iter-1 );	/**/

   /* for each pattern of the mapped space */
   for( q = 0; q < d; q++ )
     meanY[ q ] = 0.0;
   for( p = 0; p < N; p++ )
   {
     for( q = 0; q < d; q++ )
     {
       dE[ q ] = 0.0; d2E2[ q ] = 0.0;
     }
     for( j = 0; j < N; j++ )
     {
       if( j != p )
       {
         /* determine dpj_star */
         if( j < p )
           m = p * (p-1)/2 + j;
         else
           m = j * (j-1)/2 + p;
         dpj_star = distX[m];
         /* determine dpj = distance between two mapped samples */
         dpj = Euclidian_Distance( Y[p], Y[j], d );

         /* printf("Y[%d]: ",p); showFV( d, Y[p] );
         printf("Y[%d]: ",j); showFV( d, Y[j] );
         printf("iter=%d  d*[%d][%d]=%f  d[%d][%d]=%f\n",
		iter, p, j, dpj_star, p, j, dpj ); DBG; /**/

         dpj_diff = dpj_star - dpj;
         dpj_prod = dpj_star * dpj;

         /* for each dimension of the vector Y[p] calculate the delta */
         for( q = 0; q < d; q++ )	/* all dimensions of reduced space */
         {
           Y_diff = Y[p][q] - Y[j][q];
           if( dpj_prod != 0.0 )
           {
             /* 1. order derivative */
             dE[ q ] += ( dpj_diff / dpj_prod ) * Y_diff;
             /* 2. order derivative */
             d2E2[ q ] += (1.0 / dpj_prod) * ( dpj_diff - (Y_diff*Y_diff/dpj)
					* (1.0 + dpj_diff / dpj) );
           }
         }
       }
     }

     /* for each pattern the 1. and 2. order derivatives are
	calculated, now adapt each pattern by the gradient descent rule */
     /* do it for each dimension */
     /* printf("Y[%d]: ", p); showFV( d, Y[p] );
     printf(">>>>delta[%d]= ", p);	/**/
     for( q = 0; q < d; q++ )
     {
       /* the multiplication by (-2/c) causes the minus sign at dE */
       delta_pq = -dE[q] / (float)fabs((double)d2E2[q]);
       /* printf("%10.5f ", delta_pq );	/**/
	/*
		Adjust the q-th dimension of the p-th mapped vector
		following the steepest descent rule
	*/
       Y_new[p][q] =  Y[p][q] - mf * delta_pq;
       meanY[ q ] += (double)Y_new[p][q];
     }
     /* printf("  New Y[%d]: ", p );  showFV( d, Y_new[p] ); DBG;	/**/
   } /* end for each pattern */

   /* shift the mapped vectors to their centroid */
   for( p = 0; p < N; p++ )
     for( q = 0; q < d; q++ )
       Y[p][q] = Y_new[p][q] - (float)(meanY[ q ] / (double)N);

   /* calculate the mapping error */
   m = 0; err = 0.0;
   for (j = 1; j < N; j++)
   {
     for (i = 0; i < j; i++)
     {
       dij_star = distX[ m ];
       dij = Euclidian_Distance( Y[i], Y[j], d );
       /* NL; printf("Y[%d]=",i); showFV( d, Y[i] );
       printf("Y[%d]=",j); showFV( d, Y[j] );
       printf("dij_star=%f dij=%f",
		dij_star, dij ); DBG;	/**/
       dij_diff = dij_star - dij;
       err += (double)(dij_diff * dij_diff / dij_star);
       m++;
     }
   }
   err /= c;
   fprintf(stdout, "Mapping error: %15.10lf      \r", err ); /**/
   fflush(stdout); /**/
 }	/* next iteration */

 FREE( distX ); FREE( dE ); FREE( d2E2 ); FREE( meanY );
 matrix_free( &Y_new, N );
 return( TRUE );
}





#define NR_ITER_MAX 32000
#define NR_ITER_DEFAULT 100

#define DIM_DEFAULT 2
static int dim = DIM_DEFAULT;
static int nrIter = NR_ITER_DEFAULT;

/*
 * Set the axes labels for the plot of Gnuplot
*/
void axes_gnu( D, plot_dim, direct_plot )
int D, plot_dim;
boolean direct_plot;
{
 if( D == 2 || D == 3 )
 {
   if( direct_plot )
   {
     sprintf( str_gnu, "First feature = %d", U->FSV[0].rank+1 );
     if( feat_description )
     {
       sprintf( str_gnu1, " (%s)", feature_desc[U->FSV[0].rank] );
       strcat( str_gnu, str_gnu1 );
     }
     GNU_xlabel( str_gnu );

     sprintf( str_gnu, "Second feature = %d", U->FSV[1].rank+1 );
     if( feat_description )
     {
       sprintf( str_gnu1, " (%s)", feature_desc[U->FSV[1].rank] );
       strcat( str_gnu, str_gnu1 );
     }
     GNU_ylabel( str_gnu );

     if( plot_dim == 3 )
     {
       sprintf( str_gnu, "Third feature = %d", U->FSV[2].rank+1 );
       if( feat_description )
       {
         sprintf( str_gnu1, " (%s)", feature_desc[U->FSV[2].rank] );
         strcat( str_gnu, str_gnu1 );
       }
       GNU_zlabel( str_gnu );
     }
   }
 }

 if( direct_plot )
 {
   sprintf( str_gnu, "%d-D PLOT OF %s", plot_dim, U->name );
   GNU_title( str_gnu );
 }
 else
 {
   sprintf( str_gnu,"SAMMON PLOT of %s: %d-D mapped to %d-D",
		U->name, D, plot_dim );
   GNU_title( str_gnu );
   GNU_xlabel( "Mapped dimension #1" );
   GNU_ylabel( "Mapped dimension #2" );
   if( plot_dim == 3 )
     GNU_zlabel( "Mapped dimension #3" );
 }
}


void get_plot_dim( D, pd )
int D, *pd;
{
 boolean direct_plot = TRUE;
 int plot_dim;

 if( D == 2 || D == 3 )
 {
   plot_dim = 2;
   if( D == 3 )
   {
     printf("Plot the 3-dimensional data directly?");
     get_boolean( direct_plot, &direct_plot );
     if( direct_plot )
       plot_dim = 3;
     else
       printf("Sammon plot: Mapping from 3-D to 2-D\n");
   }
   else	/* D = 2 */
     direct_plot = TRUE;

   if( direct_plot )
   {
     printf("Number of selected features is %d\n", D );
     printf("Plotting the data directly in %d dimensions,", D );
     printf(" without using the Sammon mapping.\n" );
   }
 }
 else /* D > 3 */
 {
   direct_plot = FALSE;
   printf("--- SAMMON PLOT ---\n");
   printf("\tMap to dimension 2 or 3? %d\b", dim );
   get_d_range( &dim, 2, 3, LEFT_CLOSED__RIGHT_CLOSED );
   plot_dim = dim;
 }

 *pd = plot_dim;
}


static void get_learn_param()
{
 printf("\tNumber of learning steps (actual=%d)? ", nrIter );
 get_d_range( &nrIter, 1, NR_ITER_MAX, LEFT_CLOSED__RIGHT_CLOSED );

 printf("\tLearning rate of Steepest Descent (actual=%4.2f)? ", MF );
 get_f_range( &MF, 0.0, MF_MAX, LEFT_OPEN__RIGHT_OPEN );
}


void sammon_plot()
{
 int i, j, k, row, N, D, plot_dim;
 boolean ok = FALSE, direct_plot = TRUE;
 float x = 0, y = 0, z = 0;  /* Initialize because of the floating exception */
 float **Samples = NULL;
 float **Y = NULL;

 if( U->nrClass == 0 )
   { printf(" Please load universe first !..."); WAIT; return; }
 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); WAIT; return; }
 if( U->nrSelFeat == 1 )
 {
   printf("No sense to map unidimensional data to 2 dimensions!..."); WAIT; return;
 }


 N = U->sumSampl;		/* all samples */
 D = U->nrSelFeat;		/* original dimension */

 /* Load samples with selected features only in buffer 'Samples' */
 Samples = matrix_alloc( N, D );
 if( Samples == NULL )
   return;

 GNU_init_gnu_file();

 get_plot_dim( D, &plot_dim );

 direct_plot = D == plot_dim;

 axes_gnu( D, plot_dim, direct_plot );

 GNU_scatter_init( U->nrClass, U->sumSampl, plot_dim );
 row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   GNU_scatter_set_name( U->C[i].name );
   GNU_scatter_set_smp( i, U->C[i].numSampl );
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     for( k = 0; k < D; k++ )
       Samples[row][k] = U->C[i].S[j*U->nrFeat+U->FSV[k].rank];
     GNU_scatter_set_index( row, i );
     row++;
   }
 }

 if( direct_plot )
 {
   for( i = 0; i < N; i++ )
   {
     x = Samples[i][0]; y = Samples[i][1]; if( plot_dim==3 ) z = Samples[i][2];
     GNU_scatter_set_point( x, y, z, i, plot_dim );
   }
 }
 else
 {
   get_learn_param();

   Y = matrix_alloc( N, plot_dim );
   if( Y == NULL )
   {
     matrix_free( &Samples, U->sumSampl );
     return;
   }
   ok = map_sammon( Samples, N, D, Y, plot_dim, nrIter, MF ); NL;

   if( ok )
   {
     for( i = 0; i < N; i++ )
     {
       x = Y[i][0]; y = Y[i][1]; if( plot_dim == 3 ) z = Y[i][2];
       GNU_scatter_set_point( x, y, z, i, plot_dim );
     }
   }
   else
   {
     printf("Unable to generate Sammon plot!"); WAIT;
   }
   matrix_free( &Y, N );
 }
 if( direct_plot || ok )
 {
   GNU_scatter_plot();
   GNU_execute();
 }
 else
 {
   GNU_close_gnu_file();
   GNU_scatter_close();
 }
 matrix_free( &Samples, U->sumSampl );
}
